import matplotlib.pyplot as plt
import numpy as np
import os
from HelperFunctions import QSTAnalysis
from HelperFunctions import MonteCarlo
from HelperFunctions import load_columns_from_file
from HelperFunctions import find_between_strings
from HelperFunctions import generate_centered_x_axis
from HelperFunctions import find_nearest
from HelperFunctions import barplot_3d
from HelperFunctions import label_4t4_matrix
from HelperFunctions import load_chi
from HelperFunctions import get_errorcount_simulation



""""Plotting for figure 2"""


def plot_dispersion_param(show=True, hcf_color='blue', smf_color='red', labelsize=12):
    data_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'Data', 'Dispersionparameter_groupindex',
                             'converted_to_csv')
    path_hcf = os.path.join(data_path, 'NANF__D__and__n_g_no_comment.csv')
    wl_hcf, disp_hcf = load_columns_from_file(path=path_hcf, col_num=2, delimiter=';', no_of_headers=1)

    path_smf = os.path.join(data_path, 'SMF28_dispersion_southhampton.csv')
    wl_smf, disp_smf_max, disp_smf_min = load_columns_from_file(path=path_smf, col_num=3, delimiter=';', no_of_headers=3)
    disp_smf_mean = (disp_smf_min + disp_smf_max) / 2
    wl, smf_start_idx = find_nearest(wl_smf, wl_hcf[0])
    wl, hcf_stop_idx = find_nearest(wl_hcf, wl_smf[-1])

    plt.plot(wl_smf[smf_start_idx:], disp_smf_mean[smf_start_idx:], color=smf_color, label='SMF28')
    plt.plot(wl_hcf[:hcf_stop_idx], disp_hcf[:hcf_stop_idx], color=hcf_color, label='HCF')
    plt.xlabel('Wavelength [nm]', fontsize=labelsize)
    plt.ylabel(r'D $[\frac{ps}{nm \cdot km}]$', fontsize=labelsize)    # Disp. param. D $[\frac{ps}{nm \cdot km}]
    plt.legend()
    plt.ylim(bottom=0)
    if show:
        plt.show()


def plot_group_index_n(show=True, hcf_color='blue',  smf_color='red', labelsize=12):
    data_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'Data', 'Dispersionparameter_groupindex',
                             'converted_to_csv')
    path_hcf = os.path.join(data_path, 'NANF__D__and__n_g_no_comment.csv')
    wl_hcf, dummy1, dummy2, ng_hcf = load_columns_from_file(path=path_hcf, col_num=4, delimiter=';', no_of_headers=1)

    path_smf = os.path.join(data_path, 'Group_index_SMF28e_no_comments.csv')
    wl_smf, ng_smf = load_columns_from_file(path=path_smf, col_num=2, delimiter=';', no_of_headers=1)
    wl_smf = wl_smf * 1e3
    wl_hcf = wl_hcf[:-1]
    ng_hcf = ng_hcf[:-1]
    wl, smf_start_idx = find_nearest(wl_smf, wl_hcf[0])
    wl, smf_stop_idx = find_nearest(wl_smf, wl_hcf[-1])

    plt.plot(wl_smf[smf_start_idx:smf_stop_idx], ng_smf[smf_start_idx:smf_stop_idx], color=smf_color, label='SMF28')
    plt.plot(wl_hcf, ng_hcf, color=hcf_color, label='HCF')
    plt.legend()
    plt.xlabel('Wavelength [nm]', fontsize=labelsize)
    plt.ylabel(r'$n_g$', fontsize=labelsize)   # Refractive index $n_g$
    if show:
        plt.show()


def plot_figure_2():
    """
    Plot arrival time with insert plot of histograms.
    :return:
    """
    # Path handling for data import
    data_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'Data', 'Histograms')
    smf_path = os.path.join(data_path, 'SMF_2023_01_12', 'Agiltron_pos_hpath1_vpath207',
                            'SMF28_Histo_start3_stop2_startdelay38590000_Hagil1_Vagil207.txt')  # 267
    hcf_path = os.path.join(data_path, 'HCF_2023_01_16', 'Agiltron_pos_hpath1_vpath207',
                            'HCF_Histo_start3_stop2_startdelay25480000_Hagil1_Vagil207.txt')  # 267
    # Import data
    smf_x, smf_y = load_columns_from_file(path=smf_path, col_num=2)
    hcf_x, hcf_y = load_columns_from_file(path=hcf_path, col_num=2)
    # Data handling for correct plotting
    smf_startdelay = find_between_strings(s=smf_path, start='_startdelay', end='_Hagil')
    smf_x_center = generate_centered_x_axis(smf_x, smf_y)
    hcf_startdelay = find_between_strings(s=hcf_path, start='_startdelay', end='_Hagil')
    hcf_x_center = generate_centered_x_axis(hcf_x, hcf_y)
    # Set figure properties
    fig = plt.figure(figsize=[18, 14])
    plt.subplots_adjust(left=0.08, right=0.97, wspace=1.5, hspace=0.8)

    # Plot subplot figure 2a
    ax = fig.add_subplot(3, 6, (1, 12))
    # Parameters for insert plots
    inset_x_size, inset_y_size = 0.25, 0.3   # Size of inset
    smfinset_x_loc, smfinset_y_loc = 0.1, 0.5 + 0.06 + 0.1      # Location of SMF inset
    hcfinset_x_loc, hcfinset_y_loc = 0.1, 0.05 + 0.06 + 0.057   # Location of HCF inset
    # Line colors
    color_hcf = 'darkblue'
    color_smf = 'firebrick'
    # fontsize handling
    example_dt = 140    # Histogram dt spacing
    latency = (float(smf_startdelay) + smf_x[np.argmax(smf_y)]) - (float(hcf_startdelay) + hcf_x[np.argmax(hcf_y)])
    ticklabel_size = 33
    axislabel_size = 38
    mainlegend_size = 26
    insetlegend_size = 23
    abc_fontsize = 45
    insert_ticklabel_size = 24

    # Normalize data
    smf_max = np.max(smf_y)
    hcf_max = np.max(hcf_y)
    # Plot figure 2a
    ax.text(x=-2, y=1.08, s='a)', fontsize=abc_fontsize, color='grey', style='italic')
    ax.plot(0, 0, linestyle='', marker='')
    ax.plot((smf_x + float(smf_startdelay)) * 1e-6, smf_y / smf_max, '-', color=color_smf, label='7.8 km SMF28')
    ax.plot((hcf_x + float(hcf_startdelay)) * 1e-6, hcf_y / hcf_max, '-', color=color_hcf, label='7.7 km HCF')
    ax.text(x=2.65e1 - 0.3, y=(np.max(hcf_y) * 0.71) / hcf_max, s='Latency = {} µs'.format(round(latency * 1e-6, 2)),     # Plot Latency text before 1e7
                  fontsize=33, color='black', style='italic')
    ax.annotate("", xy=((float(hcf_startdelay) + hcf_x[np.argmax(hcf_y)]) * 1e-6, ((np.max(hcf_y) - 3) * 0.7) / hcf_max),    # Plot arrow under text
                xytext=((float(smf_startdelay) + smf_x[np.argmax(smf_y)]) * 1e-6, ((np.max(hcf_y) - 3) * 0.7) / hcf_max),
                arrowprops=dict(arrowstyle="<->, head_width=0.5", lw=2.5))
    # Plot SMF insert of figure 2a
    axin_smf = ax.inset_axes([smfinset_x_loc, smfinset_y_loc, inset_x_size, inset_y_size])
    axin_smf.plot(smf_x_center, smf_y / smf_max, color=color_smf, label='SMF28', zorder=1)
    axin_smf.legend(fontsize=insetlegend_size, loc=(0.525, 0.69), handlelength=0.8)
    axin_smf.set_xlabel('Time [ps]', fontsize=axislabel_size)
    # subregion of the original image
    x1, x2 = -420, 420 + 80 # axis limits
    axin_smf.set_xlim(x1, x2)   # Set axis limits in insert
    axin_smf.set_xticks([-250, 250])
    axin_smf.set_ylim(-0.43, 1.05)
    axin_smf.set_yticks([0, 0.5, 1])  # Set displayed y-ticks
    axin_smf.tick_params(axis='both', which='major', labelsize=insert_ticklabel_size)  # insert tick_label size
    #axin_smf.set_xticklabels([])  # remove x-tick label in insert
    #axin_smf.set_yticklabels([])  # remove y-tick label in insert
    # Draw dashed black lines
    axin_smf.plot([0, 0], [0, 1], '--', color='black', zorder=0, linewidth=2)
    axin_smf.plot([-137, -137], [0, 0.23], '--', color='black', zorder=0, linewidth=2)
    axin_smf.plot([137, 137], [0, 0.28], '--', color='black', zorder=0, linewidth=2)
    # Plot time-bin spacing text
    axin_smf.text(x=-120, y=-0.24 - 0.08, s=r'$\Delta t = {} ps$'.format(example_dt), fontsize=29, color='black', style='italic')
    # Make arrrow for dt
    axin_smf.annotate("", xy=(-145, -0.05), xytext=(0, -0.05),
                      arrowprops=dict(arrowstyle="<|-|>", facecolor='black', lw=2.5))

    # Plot grey lines around insert positions
    x1, y1 = [1.5e1 - 0.21e1, 3.861e1], [smf_max / smf_max, 0]
    x2, y2 = [1.5e1 - 0.21e1, 3.861e1], [(111 + 14.5 + 24) / smf_max, 0]
    plt.plot(x1, y1, x2, y2, marker='', color='grey', linewidth=1, zorder=0)    # Draws line from data to insert
    # Plot HCF insert of figure 2a
    axin_hcf = ax.inset_axes([hcfinset_x_loc, hcfinset_y_loc, inset_x_size, inset_y_size])
    axin_hcf.plot(hcf_x_center, hcf_y / hcf_max, color=color_hcf, label='HCF', zorder=1)
    axin_hcf.legend(fontsize=insetlegend_size, loc=(0.63, 0.69), handlelength=0.8)
    axin_hcf.set_xlabel('Time [ps]', fontsize=axislabel_size)
    # subregion of the original image
    x1, x2 = -420, 420 + 80            # axis limits
    axin_hcf.set_xlim(x1, x2)       # Set axis limits in insert
    axin_hcf.set_xticks([-250, 250])
    axin_hcf.set_ylim(-0.43, 1.05)   # Set y-limit
    axin_hcf.set_yticks([0, 0.5, 1])     # Set displayed y-ticks
    axin_hcf.tick_params(axis='both', which='major', labelsize=insert_ticklabel_size)  # insert tick_label size
    # Draw dashed black lines
    axin_hcf.plot([0, 0], [0, 1], '--', color='black', zorder=0, linewidth=2)
    axin_hcf.plot([-137, -137], [0, 0.3], '--', color='black', zorder=0, linewidth=2)
    axin_hcf.plot([137, 137], [0, 0.18], '--', color='black', zorder=0, linewidth=2)
    # Plot time-bin spacing as text
    axin_hcf.text(x=-120, y=-0.24 - 0.08, s=r'$\Delta t = {} ps$'.format(example_dt),   # 78, 0.62
         fontsize=29, color='black', style='italic')
    # Make arrow for dt
    axin_hcf.annotate("", xy=(-145, -0.05), xytext=(0, -0.05),
                      arrowprops=dict(arrowstyle="<|-|>", facecolor='black', lw=2.5))
    # Draw grey lines to insert position
    x1, y1 = [1.5e1 - 0.21e1, 2.548e1], [(57 + 11.5 + 10.5) / hcf_max, 0]
    x2, y2 = [1.5e1 - 0.21e1, 2.548e1], [(1 + 11.5 + 10.5) / hcf_max, 0]
    plt.plot(x1, y1, x2, y2, marker='', color='grey', linewidth=1, zorder=0)    # Draws line from data to insert

    plt.xlabel('Time [µs]', fontsize=axislabel_size)
    plt.ylabel('Coincidence counts', fontsize=axislabel_size)
    plt.legend(fontsize=mainlegend_size, loc=(0.38, 0.81), handlelength=1.5)
    plt.grid(axis='x')
    plt.xticks(fontsize=ticklabel_size)
    plt.yticks(fontsize=ticklabel_size)

    # Plot figure 2b
    bx = fig.add_subplot(3, 6, (14, 15))
    bx.text(x=1525, y=23.8, s='b)', fontsize=abc_fontsize, color='grey', style='italic')
    plot_dispersion_param(show=False, hcf_color=color_hcf, smf_color=color_smf, labelsize=axislabel_size)
    bx.set_yticks([0, 4, 8, 12, 16, 20])
    plt.xticks(fontsize=ticklabel_size - 2)
    plt.yticks(fontsize=ticklabel_size - 2)
    bx.set_xticks([1550, 1600])
    bx.legend(fontsize=mainlegend_size, loc='center right', handlelength=1.5)

    # Plot figure 2c
    cx = fig.add_subplot(3, 6, (16, 17))
    cx.text(x=1525, y=1.52, s='c)', fontsize=abc_fontsize, color='grey', style='italic')
    plot_group_index_n(show=False, hcf_color=color_hcf, smf_color=color_smf, labelsize=axislabel_size)
    plt.xticks(fontsize=ticklabel_size - 2)
    plt.yticks(fontsize=ticklabel_size - 2)
    cx.set_xticks([1550, 1600])
    cx.legend(fontsize=mainlegend_size, loc='center right', handlelength=1.5)
    plt.show()


"""Plotting for figure 3"""


def get_rhos_for_figure_3():
    """
    This function generates the density matrix files for figure 3.
    :return:
    """
    source = QSTAnalysis()
    source_folder = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'Data', 'source_tomo')
    source.singles_name = os.path.join(source_folder, '2qubit_Tomo_w126ns_t10s_singles_20221110_1649.txt')
    source.coincidences_name = os.path.join(source_folder, '2qubit_Tomo_w126ns_t10s_coincidences_20221110_1649.txt')
    source.efficiencies_name = os.path.join(source_folder, '2_qubit_efficiency_measurement_221110-1115.txt')
    source_window = 0.126
    source_int_time = 10
    source.read_data_2qubit(window_used=source_window, int_time=source_int_time)
    source.load_efficiencies()
    source.run_2qubit_tomography()
    source.save_density_matrix()


    smf = QSTAnalysis()
    smf_folder = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'Data', 'dt_VS_con_pur_data', 'SMF_sigma_3')
    smf.singles_name = os.path.join(smf_folder, 'SMF_2qubitTomo_hpath1_vpathvpos_587_w161ps_mtime120ps_Singles.txt')
    smf.coincidences_name = os.path.join(smf_folder,
                                         'SMF_2qubitTomo_hpath1_vpathvpos_587_w161ps_mtime120ps_Coincidences.txt')
    smf_window = 0.161
    smf_int_time = 120
    smf.read_data_2qubit(window_used=smf_window, int_time=smf_int_time)
    smf.efficiencies = np.array([1, 1, 1, 1])
    smf.run_2qubit_tomography()
    smf.save_density_matrix()

    hcf = QSTAnalysis()
    hcf_folder = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'Data', 'dt_VS_con_pur_data', 'HCF_sigma_3')
    hcf.singles_name = os.path.join(hcf_folder, 'HCF_2qubitTomo_hpath1_vpathvpos_587_w69ps_mtime150ps_Singles.txt')
    hcf.coincidences_name = os.path.join(hcf_folder,
                                         'HCF_2qubitTomo_hpath1_vpathvpos_587_w69ps_mtime150ps_Coincidences.txt')
    hcf_window = 0.69
    hcf_int_time = 150
    hcf.read_data_2qubit(window_used=hcf_window, int_time=hcf_int_time)
    hcf.efficiencies = np.array([1, 1, 1, 1])
    hcf.run_2qubit_tomography()
    hcf.save_density_matrix()


def plot_figure_3():
    """
    Plot real and imaginary part of the density matrices for the source, after smf28 and hcf with dt = 520
    :return:
    """
    get_rhos_for_figure_3()     # Here we generate the density matrix files for figure 3
    # Density matrix measured on source with a 3 sigma window
    source = QSTAnalysis()
    source.folder = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'Data', 'source_tomo',
                                 '2qubit_Tomo_w126ns_t10s_coincidences_20221110_1649.txt_RHO.txt')
    source.read_density_matrix(rho_path=source.folder)
    # Density matrix measured after SMF28 fiber with dt = 520 ps and window = 3 sigma
    smf = QSTAnalysis()
    smf.folder = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'Data', 'dt_VS_con_pur_data', 'SMF_sigma_3',
                              'SMF_2qubitTomo_hpath1_vpathvpos_587_w161ps_mtime120ps_Coincidences.txt_RHO.txt')
    smf.read_density_matrix(rho_path=smf.folder)
    # Density matrix measured after HCF fiber with dt = 520 ps and window = 3 sigma
    hcf = QSTAnalysis()
    hcf.folder = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'Data', 'dt_VS_con_pur_data', 'HCF_sigma_3',
                              'HCF_2qubitTomo_hpath1_vpathvpos_587_w69ps_mtime150ps_Coincidences.txt_RHO.txt')
    hcf.read_density_matrix(rho_path=hcf.folder)


    # Plot params
    real_axis = r'$\Re[\rho]$'
    sub_fontsize = 23
    # Generating the subplots
    fig = plt.figure(figsize=(16, 8))

    plt.subplots_adjust(wspace=0.1, hspace=0.6)
    # Rho-Source
    # 2 qubit
    barplot_3d(matrix=source.rho_real, z_label=real_axis, figure=fig, elev_view=10, azim_view=-50,
               sub_row=3, sub_col=3, sub_idx=2, xtick_rotation=45, ytick_rotation=-45)
    fig.text(x=0.372 + 0.045, y=0.928 - 0.02, s='a) Source', fontsize=sub_fontsize, color='grey', style='italic')
    source_values = r'$\gamma_{source}$' + f' = {round(source.purity_rho, 2)}, ' + r'$C_{source}$' + f' = {round(source.concurrence_rho, 2)}'
    fig.text(x=0.372 + 0.045, y=0.89 - 0.02, s=source_values, fontsize=sub_fontsize, color='black',
              style='italic')

    # Rho-SMF
    # 2 qubit
    barplot_3d(matrix=smf.rho_real, z_label=real_axis, figure=fig, elev_view=10, azim_view=-50,
               sub_row=3, sub_col=3, sub_idx=(4, 5), xtick_rotation=45, ytick_rotation=-45)
    fig.text(x=0.372 - 0.11, y=0.928 - 0.31, s=r'b) 7.8 km SMF28, $\Delta t = 520 ps$', fontsize=sub_fontsize, color='grey', style='italic')
    smf_values = r'$\gamma_{SMF28}$' + f' = {round(smf.purity_rho, 2)}, ' + r'$C_{SMF28}$' + f' = {round(smf.concurrence_rho, 2)}'
    fig.text(x=0.372 - 0.11, y=0.89 - 0.31, s=smf_values, fontsize=sub_fontsize, color='black',
             style='italic')

    # Rho-HCF
    # 2 qubit
    barplot_3d(matrix=hcf.rho_real, z_label=real_axis, figure=fig, elev_view=10, azim_view=-50,
               sub_row=3, sub_col=3, sub_idx=(5, 6), xtick_rotation=45, ytick_rotation=-45)
    fig.text(x=0.674 - 0.11, y=0.928 - 0.31, s=r'c) 7.7 km HCF, $\Delta t = 520 ps$', fontsize=sub_fontsize, color='grey', style='italic')
    hcf_values = r'$\gamma_{HCF}$' + f' = {round(hcf.purity_rho, 2)}, ' + r'$C_{HCF}$' + f' = {round(hcf.concurrence_rho, 2)}'
    fig.text(x=0.674 - 0.11, y=0.89 - 0.31, s=hcf_values, fontsize=sub_fontsize, color='black',
             style='italic')

    # Chi-matrix Theory
    chi_th_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'Data', 'ProcessTomo',
                               'theory_depolarizing_channel_chi_matrix.txt')
    chi_th = load_chi(chi_th_file)
    abs_chi_th = abs(chi_th)
    ax = barplot_3d(matrix=abs_chi_th, z_label='', figure=fig, color='flag',
                    azim_view=-63, elev_view=22, sub_row=3, sub_col=3, sub_idx=(8, 9), ytick_rotation=3)
    label_4t4_matrix('I', 'X', 'Y', 'Z')
    ax.set_zticks([-1, -0.5, 0, 0.5, 1])
    fig.text(x=0.674 - 0.13, y=0.448 - 0.15, s='e) ' + r'$\chi_{depol}$' + ' - Depolarizing channel', fontsize=sub_fontsize,
             color='grey', style='italic')
    # Chi-matrix Experiment
    chi_exp_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'Data', 'ProcessTomo',
                                'experiment_chi_matrix.txt')
    chi_exp = load_chi(chi_exp_file)
    abs_chi_exp = abs(chi_exp)
    bx = barplot_3d(matrix=abs_chi_exp, z_label='', figure=fig, color='flag',
                    azim_view=-63, elev_view=22, sub_row=3, sub_col=3, sub_idx=(7, 8), ytick_rotation=3)
    label_4t4_matrix('I', 'X', 'Y', 'Z')
    bx.set_zticks([-1, -0.5, 0, 0.5, 1])
    fig.text(x=0.372 - 0.12, y=0.448 - 0.15, s='d) ' + r'$\chi_{HCF}$' + ' - 7.7 km HCF channel', fontsize=sub_fontsize,
             color='grey', style='italic')

    plt.show()


"""Plotting for figure 4"""


def run_monte_carlo_on_hcf_rho_for_dt_scan_tomo():
    a = QSTAnalysis()

    purity_array = np.array([])
    purity_stds = np.copy(purity_array)
    concurrence_array = np.array([])
    concurrence_stds = np.copy(concurrence_array)
    dt_array = np.arange(0, 521, 20)    # in picoseconds.

    # a.fiber = '7.8 km SMF28 fiber'
    # window = 3 * 53.686 * 1e-3  # In nanoseconds.
    # int_time = 120              # In seconds.
    a.fiber = '7.7 km HCF'
    window = 3 * 23.111 * 1e-3  # In nanoseconds.
    int_time = 150  # in seconds.

    fiber_dir = 'HCF_sigma_3'
    cc_files, s_files = a.get_singles_and_coincidence_file(fiber_dir=fiber_dir)
    a.efficiencies = np.array([1, 1, 1, 1])  # Add detector specification. 1-2, 1-4, 3-2, 3-4<#

    """"SET main MONTE CARLO SIMULATION steps"""
    mc = MonteCarlo()
    mc_runs = 1000

    for i in range(len(cc_files)):
        print(f'\nProcess file {i + 1} out of {len(cc_files)}:\n')
        a.singles_name = s_files[i]
        a.coincidences_name = cc_files[i]
        a.delta_t = dt_array[i]
        """DO INITIAL TOMOGRAPHY"""
        a.read_data_2qubit(window_used=window, int_time=int_time)
        a.run_2qubit_tomography()
        a.extract_complex_numbers()
        a.analyse_rho()
        purity_array = np.append(purity_array, a.purity_rho)
        concurrence_array = np.append(concurrence_array, a.concurrence_rho)
        """SET individual MONTE CARLO SIMULATION steps"""
        mc_purities = np.array([])
        mc_concurrence = np.array([])
        probabilities, tot_counts = mc.get_probs_from_counts(measured_counts=a.coincidences)
        """RUN MONTE CARLO SIMULATION"""
        for run in range(mc_runs):
            print(f'MC run {run + 1} out of {mc_runs}...')
            mc_counts = mc.monte_carlo_step(total_count_vector=tot_counts, probability_array=probabilities)
            a.coincidences = mc_counts

            a.run_2qubit_tomography()
            a.extract_complex_numbers()
            a.analyse_rho()
            mc_purities = np.append(mc_purities, a.purity_rho)
            mc_concurrence = np.append(mc_concurrence, a.concurrence_rho)

        std_purity = np.std(mc_purities)
        purity_stds = np.append(purity_stds, std_purity)
        std_concurrence = np.std(mc_concurrence)
        concurrence_stds = np.append(concurrence_stds, std_concurrence)
        print(f'State purity: {purity_array[i]} ± {std_purity}')
        print(f'State concurrence: {concurrence_array[i]} ± {std_concurrence}')
        print('\n\n')

    fname = 'dt_con_pur_conERR_purERR_sigma_3Measurement_2023Jan17_23h46.txt'
    a.save_concurrence_purity_dt_conerr_purerr(fiber_dir=fiber_dir,
                                               dt=dt_array, concurr=concurrence_array, purity=purity_array,
                                               con_err=concurrence_stds, pur_err=purity_stds, fname=fname)


def run_monte_carlo_on_smf_rho_for_dt_scan_tomo():
    a = QSTAnalysis()

    purity_array = np.array([])
    purity_stds = np.copy(purity_array)
    concurrence_array = np.array([])
    concurrence_stds = np.copy(concurrence_array)
    dt_array = np.arange(0, 521, 20)    # in picoseconds.

    a.fiber = '7.8 km SMF28 fiber'
    window = 3 * 53.686 * 1e-3  # In nanoseconds.
    int_time = 120              # In seconds.

    fiber_dir = 'SMF_sigma_3'
    cc_files, s_files = a.get_singles_and_coincidence_file(fiber_dir=fiber_dir)
    a.efficiencies = np.array([1, 1, 1, 1])  # Add detector specification. 1-2, 1-4, 3-2, 3-4<#

    """"SET main MONTE CARLO SIMULATION steps"""
    mc = MonteCarlo()
    mc_runs = 1000

    for i in range(len(cc_files)):
        print(f'\nProcess file {i + 1} out of {len(cc_files)}:\n')
        a.singles_name = s_files[i]
        a.coincidences_name = cc_files[i]
        a.delta_t = dt_array[i]
        """DO INITIAL TOMOGRAPHY"""
        a.read_data_2qubit(window_used=window, int_time=int_time)
        a.run_2qubit_tomography()
        a.extract_complex_numbers()
        a.analyse_rho()
        purity_array = np.append(purity_array, a.purity_rho)
        concurrence_array = np.append(concurrence_array, a.concurrence_rho)
        """SET individual MONTE CARLO SIMULATION steps"""
        mc_purities = np.array([])
        mc_concurrence = np.array([])
        probabilities, tot_counts = mc.get_probs_from_counts(measured_counts=a.coincidences)
        """RUN MONTE CARLO SIMULATION"""
        for run in range(mc_runs):
            print(f'MC run {run + 1} out of {mc_runs}...')
            mc_counts = mc.monte_carlo_step(total_count_vector=tot_counts, probability_array=probabilities)
            a.coincidences = mc_counts

            a.run_2qubit_tomography()
            a.extract_complex_numbers()
            a.analyse_rho()
            mc_purities = np.append(mc_purities, a.purity_rho)
            mc_concurrence = np.append(mc_concurrence, a.concurrence_rho)

        std_purity = np.std(mc_purities)
        purity_stds = np.append(purity_stds, std_purity)
        std_concurrence = np.std(mc_concurrence)
        concurrence_stds = np.append(concurrence_stds, std_concurrence)
        print(f'State purity: {purity_array[i]} ± {std_purity}')
        print(f'State concurrence: {concurrence_array[i]} ± {std_concurrence}')
        print('\n\n')

    fname = 'dt_con_pur_conERR_purERR_sigma_3Measurement_2023Jan12_16h25.txt'
    a.save_concurrence_purity_dt_conerr_purerr(fiber_dir=fiber_dir,
                                               dt=dt_array, concurr=concurrence_array, purity=purity_array,
                                               con_err=concurrence_stds, pur_err=purity_stds, fname=fname)


def generate_data_for_figure_4():
    """
    This function runs all tomographies for both fiber types with every time-bin spacing in figure 4. Also it performs
    the Monte Carlo simulation to estimate the errors for concurrence and purity.
    WARNING: This function takes several hours to run. The resulting data is already generated for plotting of figure 4.
    :return:
    """
    run_monte_carlo_on_smf_rho_for_dt_scan_tomo()
    run_monte_carlo_on_hcf_rho_for_dt_scan_tomo()


def plot_figure_4():
    """
    This function plots the concurrence and purity data together with the simulations plus density matrices
    for both fibers at an example time spacing.
    :return:
    """
    # SMF-Data
    smf_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'Data', 'dt_VS_con_pur_data', 'SMF_sigma_3',
                              'dt_con_pur_conERR_purERR_sigma_3Measurement_2023Jan12_16h25.txt')
    smf_dt, smf_con, smf_pur, smf_con_std, smf_pur_std = load_columns_from_file(path=smf_path, col_num=5)

    # HCF-Data
    hcf_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'Data', 'dt_VS_con_pur_data', 'HCF_sigma_3',
                              'dt_con_pur_conERR_purERR_sigma_3Measurement_2023Jan17_23h46.txt')
    hcf_dt, hcf_con, hcf_pur, hcf_con_std, hcf_pur_std = load_columns_from_file(path=hcf_path, col_num=5)

    # Density matrix measured after SMF28 fiber with dt = ... ps and window = 3 sigma
    example_dt = 140
    smf = QSTAnalysis()
    smf.folder = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'Data', 'dt_VS_con_pur_data', 'SMF_sigma_3',
                              'SMF_2qubitTomo_hpath1_vpathvpos_{}_w161ps_mtime120ps_Coincidences.txt_RHO.txt'.format(example_dt + 67))
    smf.read_density_matrix(rho_path=smf.folder)
    # Initial SMF28 density matrix at dt = 520 ps
    init_smf = QSTAnalysis()
    init_smf.folder = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'Data', 'dt_VS_con_pur_data', 'SMF_sigma_3',
                                   'SMF_2qubitTomo_hpath1_vpathvpos_587_w161ps_mtime120ps_Coincidences.txt_RHO.txt')
    init_smf.read_density_matrix(rho_path=init_smf.folder)
    # Density matrix measured after HCF fiber with dt = ... ps and window = 3 sigma
    hcf = QSTAnalysis()
    hcf.folder = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'Data', 'dt_VS_con_pur_data', 'HCF_sigma_3',
                              'HCF_2qubitTomo_hpath1_vpathvpos_{}_w69ps_mtime150ps_Coincidences.txt_RHO.txt'.format(example_dt + 67))
    hcf.read_density_matrix(rho_path=hcf.folder)
    # Initial HCF density matrix at dt = 520 ps
    init_hcf = QSTAnalysis()
    init_hcf.folder = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'Data', 'dt_VS_con_pur_data', 'HCF_sigma_3',
                                   'HCF_2qubitTomo_hpath1_vpathvpos_587_w69ps_mtime150ps_Coincidences.txt_RHO.txt')
    init_hcf.read_density_matrix(rho_path=init_hcf.folder)

    fig = plt.figure(figsize=(16, 8))

    # Plot params
    real_axis = r'$\Re[\rho]$'
    img_axis = r'$\Im[\rho]$'
    con_pur_dt_labelsize = 17
    con_pur_dt_ticksize = 14
    abc_fontsize = 30
    color_hcf = 'darkblue'
    color_smf = 'firebrick'
    rhotitle_size = 25
    rho_titel_x = -0.3
    rho_titel_y = -0.55
    # Error params
    errorcap_size = 5
    errorbar_thickness = 1.5

    plt.subplots_adjust(wspace=0.9, hspace=0.1)
    # Concurrence, purity VS dt
    #plt.subplot(211)
    ax = fig.add_subplot(2, 1, 1)
    # SMF28
    ax.text(x=-55, y=1.15, s='a)', fontsize=abc_fontsize, color='grey', style='italic')
    ax.plot(smf_dt, smf_con, marker='s', linestyle='', color=color_smf, label=r'$C(\rho)_{SMF28}$', zorder=2)
    ax.errorbar(x=smf_dt, y=smf_con, yerr=smf_con_std, fmt='none', ecolor='k', capsize=errorcap_size,
                 elinewidth=errorbar_thickness, capthick=errorbar_thickness, markersize=1)
    ax.plot(smf_dt, smf_pur, marker='v', linestyle='', color=color_smf, label=r'$\gamma(\rho)_{SMF28}$', zorder=2)
    ax.errorbar(x=smf_dt, y=smf_pur, yerr=smf_pur_std, fmt='none', ecolor='k', capsize=errorcap_size,
                 elinewidth=errorbar_thickness, capthick=errorbar_thickness, markersize=1)
    sim_smf_dt, sim_smf_pur, sim_smf_con = get_errorcount_simulation(init_rho=init_smf.rho, window_no=3,
                                                                     fiber='smf')
    ax.plot(sim_smf_dt, sim_smf_con, marker='', linestyle='--', color=color_smf,
             label=r'$C(\rho)_{SMF28} simulation$', zorder=2)
    ax.plot(sim_smf_dt, sim_smf_pur, marker='', linestyle='--', color=color_smf,
             label=r'$\gamma(\rho)_{SMF28} simulation$', zorder=2)
    # HCF
    ax.plot(hcf_dt, hcf_con, marker='s', linestyle='', color=color_hcf, label=r'$C(\rho)_{HCF}$', zorder=1)
    ax.errorbar(x=hcf_dt, y=hcf_con, yerr=hcf_con_std, fmt='none', ecolor='k', capsize=errorcap_size,
                 elinewidth=errorbar_thickness, capthick=errorbar_thickness, markersize=1)
    ax.plot(hcf_dt, hcf_pur, marker='v', linestyle='', color=color_hcf, label=r'$\gamma(\rho)_{HCF}$', zorder=1)
    ax.errorbar(x=hcf_dt, y=hcf_pur, yerr=hcf_pur_std, fmt='none', ecolor='k', capsize=errorcap_size,
                 elinewidth=errorbar_thickness, capthick=errorbar_thickness, markersize=1)
    sim_hcf_dt, sim_hcf_pur, sim_hcf_con = get_errorcount_simulation(init_rho=init_hcf.rho, window_no=3,
                                                                     fiber='hcf')
    ax.plot(sim_hcf_dt, sim_hcf_con, marker='', linestyle='-', color=color_hcf,
             label=r'$C(\rho)_{HCF} simulation$',
             zorder=2)
    ax.plot(sim_hcf_dt, sim_hcf_pur, marker='', linestyle='-', color=color_hcf,
             label=r'$\gamma(\rho)_{HCF} simulation$', zorder=2)

    line_color = 'black'
    ax.axvline(x=example_dt, color=line_color, zorder=0)  # , label=r'$\Delta t = {} ps$'.format(example_dt))
    ax.text(x=example_dt + 5, y=0.1, s=r'$\Delta t = {} ps$'.format(example_dt),
             fontsize=20, color=line_color, style='italic')
    plt.xlabel(r'Time-Bin spacing $\Delta t$ [ps]', fontsize=con_pur_dt_labelsize)
    plt.ylabel(r'$C(\rho), \gamma (\rho)$', fontsize=con_pur_dt_labelsize)
    plt.grid()
    plt.xticks(fontsize=con_pur_dt_ticksize)
    plt.yticks(fontsize=con_pur_dt_ticksize)
    plt.legend(loc='lower right', fontsize='large')
    # Density matrices
    # SMF28
    plt.text(x=-55, y=-0.4, s='b)', fontsize=abc_fontsize, color='grey', style='italic')
    plt.text(x=270, y=-0.4, s='c)', fontsize=abc_fontsize, color='grey', style='italic')
    barplot_3d(matrix=smf.rho_real, z_label=real_axis, figure=fig, sub_row=2, sub_col=4, sub_idx=5, elev_view=10, azim_view=-47,
               ticksize=15)
    barplot_3d(matrix=smf.rho_imag, z_label=img_axis, figure=fig, sub_row=2, sub_col=4, sub_idx=6, elev_view=10, azim_view=-47,
               ticksize=15)
    plt.title(r'$\rho_{SMF28, \Delta t=140 ps}$', x=rho_titel_x, y=rho_titel_y, fontsize=rhotitle_size,
              color='grey',
              style='italic')
    # HCF
    barplot_3d(matrix=hcf.rho_real, z_label=real_axis, figure=fig, sub_row=2, sub_col=4, sub_idx=7, elev_view=10, azim_view=-47,
               ticksize=15)
    barplot_3d(matrix=hcf.rho_imag, z_label=img_axis, figure=fig, sub_row=2, sub_col=4, sub_idx=8, elev_view=10, azim_view=-47,
               ticksize=15)
    plt.title(r'$\rho_{HCF, \Delta t=140 ps}$', x=rho_titel_x, y=rho_titel_y, fontsize=rhotitle_size, color='grey',
              style='italic')

    plt.show()


if __name__ == '__main__':
    """Run this functions to generate the plots in the paper."""
    plot_figure_2()
    plot_figure_3()
    plot_figure_4()

